/****************************************************************************
** Meta object code from reading C++ file 'kcompactdisc.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.1)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../libkcompactdisc-25.12.3/src/kcompactdisc.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'kcompactdisc.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.1. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN12KCompactDiscE_t {};
} // unnamed namespace

template <> constexpr inline auto KCompactDisc::qt_create_metaobjectdata<qt_meta_tag_ZN12KCompactDiscE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "KCompactDisc",
        "playoutPositionChanged",
        "",
        "position",
        "playoutTrackChanged",
        "track",
        "randomPlaylistChanged",
        "loopPlaylistChanged",
        "discChanged",
        "tracks",
        "discInformation",
        "KCompactDisc::DiscInfo",
        "info",
        "discStatusChanged",
        "KCompactDisc::DiscStatus",
        "status",
        "volumeChanged",
        "volume",
        "balanceChanged",
        "balance",
        "playTrack",
        "playPosition",
        "play",
        "next",
        "prev",
        "pause",
        "stop",
        "eject",
        "loop",
        "random",
        "doCommand",
        "KCompactDisc::DiscCommand",
        "metadataLookup",
        "setRandomPlaylist",
        "setLoopPlaylist",
        "setAutoMetadataLookup",
        "setVolume",
        "setBalance"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'playoutPositionChanged'
        QtMocHelpers::SignalData<void(unsigned int)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::UInt, 3 },
        }}),
        // Signal 'playoutTrackChanged'
        QtMocHelpers::SignalData<void(unsigned int)>(4, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::UInt, 5 },
        }}),
        // Signal 'randomPlaylistChanged'
        QtMocHelpers::SignalData<void(bool)>(6, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'loopPlaylistChanged'
        QtMocHelpers::SignalData<void(bool)>(7, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Signal 'discChanged'
        QtMocHelpers::SignalData<void(unsigned int)>(8, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::UInt, 9 },
        }}),
        // Signal 'discInformation'
        QtMocHelpers::SignalData<void(KCompactDisc::DiscInfo)>(10, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 11, 12 },
        }}),
        // Signal 'discStatusChanged'
        QtMocHelpers::SignalData<void(KCompactDisc::DiscStatus)>(13, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 14, 15 },
        }}),
        // Signal 'volumeChanged'
        QtMocHelpers::SignalData<void(unsigned int)>(16, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::UInt, 17 },
        }}),
        // Signal 'balanceChanged'
        QtMocHelpers::SignalData<void(unsigned int)>(18, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::UInt, 19 },
        }}),
        // Slot 'playTrack'
        QtMocHelpers::SlotData<void(unsigned int)>(20, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::UInt, 5 },
        }}),
        // Slot 'playPosition'
        QtMocHelpers::SlotData<void(unsigned int)>(21, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::UInt, 3 },
        }}),
        // Slot 'play'
        QtMocHelpers::SlotData<void()>(22, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'next'
        QtMocHelpers::SlotData<void()>(23, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'prev'
        QtMocHelpers::SlotData<void()>(24, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'pause'
        QtMocHelpers::SlotData<void()>(25, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'stop'
        QtMocHelpers::SlotData<void()>(26, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'eject'
        QtMocHelpers::SlotData<void()>(27, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'loop'
        QtMocHelpers::SlotData<void()>(28, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'random'
        QtMocHelpers::SlotData<void()>(29, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'doCommand'
        QtMocHelpers::SlotData<void(KCompactDisc::DiscCommand)>(30, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 31, 2 },
        }}),
        // Slot 'metadataLookup'
        QtMocHelpers::SlotData<void()>(32, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'setRandomPlaylist'
        QtMocHelpers::SlotData<void(bool)>(33, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'setLoopPlaylist'
        QtMocHelpers::SlotData<void(bool)>(34, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'setAutoMetadataLookup'
        QtMocHelpers::SlotData<void(bool)>(35, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 2 },
        }}),
        // Slot 'setVolume'
        QtMocHelpers::SlotData<void(unsigned int)>(36, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::UInt, 17 },
        }}),
        // Slot 'setBalance'
        QtMocHelpers::SlotData<void(unsigned int)>(37, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::UInt, 19 },
        }}),
    };
    QtMocHelpers::UintData qt_properties {
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<KCompactDisc, qt_meta_tag_ZN12KCompactDiscE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject KCompactDisc::staticMetaObject = { {
    QMetaObject::SuperData::link<QObject::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN12KCompactDiscE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN12KCompactDiscE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN12KCompactDiscE_t>.metaTypes,
    nullptr
} };

void KCompactDisc::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<KCompactDisc *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->playoutPositionChanged((*reinterpret_cast<std::add_pointer_t<uint>>(_a[1]))); break;
        case 1: _t->playoutTrackChanged((*reinterpret_cast<std::add_pointer_t<uint>>(_a[1]))); break;
        case 2: _t->randomPlaylistChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 3: _t->loopPlaylistChanged((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 4: _t->discChanged((*reinterpret_cast<std::add_pointer_t<uint>>(_a[1]))); break;
        case 5: _t->discInformation((*reinterpret_cast<std::add_pointer_t<KCompactDisc::DiscInfo>>(_a[1]))); break;
        case 6: _t->discStatusChanged((*reinterpret_cast<std::add_pointer_t<KCompactDisc::DiscStatus>>(_a[1]))); break;
        case 7: _t->volumeChanged((*reinterpret_cast<std::add_pointer_t<uint>>(_a[1]))); break;
        case 8: _t->balanceChanged((*reinterpret_cast<std::add_pointer_t<uint>>(_a[1]))); break;
        case 9: _t->playTrack((*reinterpret_cast<std::add_pointer_t<uint>>(_a[1]))); break;
        case 10: _t->playPosition((*reinterpret_cast<std::add_pointer_t<uint>>(_a[1]))); break;
        case 11: _t->play(); break;
        case 12: _t->next(); break;
        case 13: _t->prev(); break;
        case 14: _t->pause(); break;
        case 15: _t->stop(); break;
        case 16: _t->eject(); break;
        case 17: _t->loop(); break;
        case 18: _t->random(); break;
        case 19: _t->doCommand((*reinterpret_cast<std::add_pointer_t<KCompactDisc::DiscCommand>>(_a[1]))); break;
        case 20: _t->metadataLookup(); break;
        case 21: _t->setRandomPlaylist((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 22: _t->setLoopPlaylist((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 23: _t->setAutoMetadataLookup((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1]))); break;
        case 24: _t->setVolume((*reinterpret_cast<std::add_pointer_t<uint>>(_a[1]))); break;
        case 25: _t->setBalance((*reinterpret_cast<std::add_pointer_t<uint>>(_a[1]))); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (KCompactDisc::*)(unsigned int )>(_a, &KCompactDisc::playoutPositionChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (KCompactDisc::*)(unsigned int )>(_a, &KCompactDisc::playoutTrackChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (KCompactDisc::*)(bool )>(_a, &KCompactDisc::randomPlaylistChanged, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (KCompactDisc::*)(bool )>(_a, &KCompactDisc::loopPlaylistChanged, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (KCompactDisc::*)(unsigned int )>(_a, &KCompactDisc::discChanged, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (KCompactDisc::*)(KCompactDisc::DiscInfo )>(_a, &KCompactDisc::discInformation, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (KCompactDisc::*)(KCompactDisc::DiscStatus )>(_a, &KCompactDisc::discStatusChanged, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (KCompactDisc::*)(unsigned int )>(_a, &KCompactDisc::volumeChanged, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (KCompactDisc::*)(unsigned int )>(_a, &KCompactDisc::balanceChanged, 8))
            return;
    }
}

const QMetaObject *KCompactDisc::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *KCompactDisc::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN12KCompactDiscE_t>.strings))
        return static_cast<void*>(this);
    return QObject::qt_metacast(_clname);
}

int KCompactDisc::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QObject::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 26)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 26;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 26)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 26;
    }
    return _id;
}

// SIGNAL 0
void KCompactDisc::playoutPositionChanged(unsigned int _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void KCompactDisc::playoutTrackChanged(unsigned int _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void KCompactDisc::randomPlaylistChanged(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void KCompactDisc::loopPlaylistChanged(bool _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1);
}

// SIGNAL 4
void KCompactDisc::discChanged(unsigned int _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1);
}

// SIGNAL 5
void KCompactDisc::discInformation(KCompactDisc::DiscInfo _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 5, nullptr, _t1);
}

// SIGNAL 6
void KCompactDisc::discStatusChanged(KCompactDisc::DiscStatus _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 6, nullptr, _t1);
}

// SIGNAL 7
void KCompactDisc::volumeChanged(unsigned int _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 7, nullptr, _t1);
}

// SIGNAL 8
void KCompactDisc::balanceChanged(unsigned int _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 8, nullptr, _t1);
}
QT_WARNING_POP
